import React from 'react';
import shallowEqual from 'shallowequal';
import objectHash from 'object-hash';
import LRU from 'lru-cache';

// copied from Apollo https://www.apollographql.com/docs/react/basics/queries.html#graphql-config-options-fetchPolicy

/**
 * Try the cache first. If match is found return that. Otherwise, send out a network request and cache it for next time.
 */
const CacheFirst = 'cache-first';

/**
 * Try the cache first but send out a network request anyway. Update component and cache if there's new data.
 */
const CacheAndNetwork = 'cache-and-network';

/**
 * Don't use the cache.
 */
const NetworkOnly = 'network-only';

/**
 * Only read from the cache. Do not send any network requests. How do you populate the cache? Nobody knows.
 */
const CacheOnly = 'cache-only';

var FetchPolicy = Object.freeze({
	CacheFirst: CacheFirst,
	CacheAndNetwork: CacheAndNetwork,
	NetworkOnly: NetworkOnly,
	CacheOnly: CacheOnly
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

function _objectWithoutProperties(obj, keys) { var target = {}; for (var i in obj) { if (keys.indexOf(i) >= 0) continue; if (!Object.prototype.hasOwnProperty.call(obj, i)) continue; target[i] = obj[i]; } return target; }

class AjaxLoader {

    constructor({
        endpoint,
        cache,
        hash = objectHash,
        batchSize = 4,
        minDelay = 8,
        maxDelay = 32,
        fetchOptions,
        refreshAllProp,

        defaultDataProp = 'ajaxData',
        defaultLoadingProp = 'ajaxLoading',
        defaultErrorProp = 'ajaxError',
        defaultEqualityCheck = shallowEqual,
        defaultHandler = setStateHandler,
        defaultFetchPolicy = CacheAndNetwork
    }) {
        this._push = requests => {
            for (let req of requests) {
                let key = this.options.hash([req.route, req.data]);

                if (this.options.cache && req.fetchPolicy !== NetworkOnly && !req._noCache) {
                    let res = this.options.cache.get(key);
                    if (res !== undefined) {
                        success(req, res.payload);
                        if (req.fetchPolicy !== CacheAndNetwork) {
                            continue;
                        }
                        req._etag = res.etag; // fixme: is it safe to mutate the request like this?
                    }
                }

                // TODO: unbatching. if a new request comes in with the same _id as an already-queued (in-batch) request, we should
                // replace that one instead of adding both.

                // console.log(JSON.stringify(map2obj(this.batch),null,2));
                // this.batch.forEach((breqs,bkey) => {
                //
                //    
                //     let oldRequests = breqs.filter(r => r._id === req._id);
                //    
                //     if(oldRequests.length) {
                //         for(let or of oldRequests) {
                //             if(or.loadingProp) {
                //                 or._component.setState(state => ({
                //                     [or.loadingProp]: state[or.loadingProp] - 1,
                //                 }));
                //             }
                //         }
                //         // filterInPlace(breqs, r => r._id !== req._id);
                //     }
                //    
                //     // let removed = 
                //     // if(removed && req.loadingProp) {
                //     //     console.log('cancelled',removed);
                //     //     req._component.setState(state => ({
                //     //         [req.loadingProp]: state[req.loadingProp] - removed,
                //     //     }));
                //     // }
                // });

                let entry = this.batch.get(key);
                if (entry) {
                    entry.push(req);
                } else {
                    this.batch.set(key, [req]);
                }

                if (req.loadingProp) {
                    // if there's a cache hit but fetch policy is cache-and-network, then....should we show the loading or not? <-- I think so. Use it to display a spinner without blocking the entire UI to indicate we're fetching fresh data. 
                    req._component.setState(state => ({
                        [req.loadingProp]: state[req.loadingProp] ? state[req.loadingProp] + 1 : 1
                    }));
                }
            }

            if (this.batch.size >= this.options.batchSize) {
                // TODO: if batch size is *exceeded* should we split the batch?
                this._run();
            } else if (this.start) {
                // if the timer has been started...
                let elapsed = performance.now() - this.start;
                if (elapsed >= this.options.maxDelay) {
                    // if max delay is exceeded, send the batch immediately
                    this._run();
                } else {
                    // otherwise, restart the timer
                    clearTimeout(this.timer);
                    this.timer = setTimeout(this._run, Math.min(this.options.minDelay, this.options.maxDelay - elapsed));
                }
            } else {
                // otherwise start the timer and queue the execution
                this.start = performance.now();
                this.timer = setTimeout(this._run, this.options.minDelay);
            }
        };

        this._run = () => {
            clearTimeout(this.timer);
            this.start = null;
            this.timer = null;
            this._send();
            this.batch.clear();
        };

        this._send = () => {
            let batchIdx = 0;
            let batch = [];
            let reqData = [];
            let keyLookup = Object.create(null);
            let rank = ++this.rankCounter;

            this.batch.forEach((reqs, key) => {
                if (!reqs.length) {
                    return;
                }

                batch[batchIdx] = reqs;
                keyLookup[batchIdx] = key;

                for (let req of reqs) {
                    this.rankLookup[req._id] = rank;
                }

                reqData[batchIdx] = {
                    route: reqs[0].route,
                    data: reqs[0].data
                };

                if (reqs[0]._etag) {
                    reqData[batchIdx].etag = reqs[0]._etag;
                }

                ++batchIdx;
            });

            // let reqData = batch.filter(reqs => reqs.length).map((reqs,key) => {
            //     let counter = this.pending[key] = this.pending[key] ? this.pending[key] + 1 : 1; 
            //     console.log(key,counter);
            //    
            //     return {
            //         route: reqs[0].route,
            //         data: reqs[0].data,
            //         counter,
            //     };
            // });

            // console.log(reqData);

            let _ref = resolveValue(this.options.fetchOptions) || {},
                { headers } = _ref,
                options = _objectWithoutProperties(_ref, ['headers']);

            // console.log('send',this.endpoint,reqData);

            fetch(this.options.endpoint, _extends({
                method: 'POST',
                credentials: 'same-origin'
            }, options, {
                headers: _extends({
                    'Accept': 'application/json, text/plain, */*',
                    'Content-Type': 'application/json'
                }, headers),
                body: JSON.stringify({
                    rank: rank,
                    requests: reqData
                })
            })).then(res => res.json()).then(fullResponse => {
                const {
                    rank: resRank,
                    responses
                } = fullResponse;

                if (reqData.length !== responses.length) {
                    throw new Error(`Server error: response length (${responses.length}) does not match request length (${reqData.length})`);
                }
                for (let i = 0; i < responses.length; ++i) {
                    let res = responses[i];

                    if (this.options.cache && res.type === 'success') {
                        this.options.cache.set(keyLookup[i], {
                            payload: res.payload,
                            etag: res.etag
                        });
                    }

                    for (let req of batch[i]) {
                        let expectedRank = this.rankLookup[req._id];

                        if (!resRank || resRank == expectedRank) {
                            switch (res.type) {
                                case 'nochange':
                                    // payload is exactly the same as we had cached, do nothing
                                    break;
                                case 'success':
                                    success(req, res.payload);
                                    break;
                                case 'error':
                                    if (process.env.NODE_ENV !== 'production') {
                                        console.group(`Error in response to route "${req.route}"`);
                                        console.error(res.payload.message);
                                        console.info("Request:", req);
                                        console.info("Response:", res.payload);
                                        console.groupEnd();
                                    }

                                    if (req.errorProp) {
                                        req._component.setState({
                                            [req.errorProp]: res.payload
                                        });
                                    }
                                    break;
                                default:
                                    throw new Error(`Server error: unexpected response type "${res.type}"`);
                            }
                        } else {
                            console.group(`Got stale response to route "${req.route}"`);
                            console.warn(`Expected ${expectedRank}, got ${resRank}`);
                            console.info("Request:", req);
                            console.info("Response:", res);
                            console.groupEnd();
                        }
                        if (req.loadingProp) {
                            req._component.setState(state => {
                                if (state[req.loadingProp] > 0) {
                                    return {
                                        [req.loadingProp]: state[req.loadingProp] ? state[req.loadingProp] - 1 : 0
                                    };
                                }
                                return {};
                            });
                        }
                    }
                }
            });
        };

        this.options = {
            endpoint,
            cache,
            hash,
            batchSize,
            minDelay,
            maxDelay,
            fetchOptions,
            refreshAllProp,
            defaultDataProp,
            defaultLoadingProp,
            defaultErrorProp,
            defaultEqualityCheck,
            defaultHandler,
            defaultFetchPolicy
        };

        this.batch = new Map();
        this.start = null;
        this.timer = null;
        this.reqId = 0;
        this.rankCounter = 0;
        this.rankLookup = Object.create(null);

        if (this.options.batchSize <= 0) {
            throw new Error(`batchSize must be > 0, got ${this.options.batchSize}`);
        }
    }

    hoc(...requests) {
        const loader = this;

        for (let req of requests) {
            setDefaults(req, {
                equalityCheck: this.options.defaultEqualityCheck,
                handler: this.options.defaultHandler,
                loadingProp: this.options.defaultLoadingProp,
                errorProp: this.options.defaultErrorProp,
                dataProp: this.options.defaultDataProp,
                refreshProp: null,
                fetchPolicy: this.options.defaultFetchPolicy,
                initialData: undefined
            });
        }

        return function ajaxLoaderEnhancer(BaseComponent) {

            class AjaxEnhanced extends React.Component {

                constructor(props) {
                    super(props);
                    // *copy* all the requests into the component
                    this.requests = requests.map(req => _extends({}, req, {
                        _component: this,
                        _id: ++loader.reqId,
                        route: resolveValue.call(this, req.route, props)
                    }));
                    this.lastData = Object.create(null);
                    this.state = Object.create(null); // fixes "Cannot read property 'loading' of null" in React 16
                }

                UNSAFE_componentWillMount() {
                    let requests = this.requests.reduce((acc, req) => {
                        let data = resolveValue.call(this, req.initialData, this.props);

                        if (data !== undefined) {
                            success(req, data);
                        } else {
                            if (typeof req.data === 'function') {
                                let data = req.data.call(this, this.props);
                                this.lastData[req._id] = data;
                                req = _extends({}, req, { data });
                            }

                            acc.push(req);
                        }

                        return acc;
                    }, []);

                    if (requests.length) {
                        loader._push(requests);
                    }
                }

                UNSAFE_componentWillReceiveProps(nextProps) {
                    let updated = this.requests.reduce((acc, req) => {
                        if (typeof req.data === 'function') {
                            let data = req.data.call(this, nextProps);
                            if (!req.equalityCheck(this.lastData[req._id], data)) {
                                this.lastData[req._id] = data;
                                acc.push(_extends({}, req, { data }));
                            }
                        }
                        return acc;
                    }, []);

                    if (updated.length) {
                        loader._push(updated);
                    }
                }

                render() {
                    let props = _extends({}, this.props, this.state);
                    let refreshFuncs = [];
                    for (let req of this.requests) {
                        const refresh = () => {
                            req = _extends({}, req, { _noCache: true });
                            if (typeof req.data === 'function') {
                                req.data = this.lastData[req._id] = req.data.call(this, this.props);
                            }
                            loader._push([req]);
                        };

                        if (req.refreshProp) {
                            props[req.refreshProp] = refresh;
                        }

                        refreshFuncs.push(refresh);
                    }
                    if (loader.options.refreshAllProp) {
                        props[loader.options.refreshAllProp] = () => {
                            for (let fn of refreshFuncs) {
                                fn();
                            }
                        };
                    }
                    return React.createElement(BaseComponent, props);
                }
            }

            AjaxEnhanced.displayName = `ajaxLoader(${BaseComponent.displayName || BaseComponent.name || 'Component'})`;
            return AjaxEnhanced;
        };
    }

}

function success(req, payload) {
    let newState = req.handler.call(req._component, payload, req);
    if (newState !== undefined) {
        // console.log('newState', req, res);
        req._component.setState(newState);
    }
}

function setDefaults(obj, defaults) {
    for (let key of Object.keys(defaults)) {
        if (obj[key] === undefined) {
            obj[key] = defaults[key];
        }
    }
}

// function map2obj(map) {
//     return Array.from(map).reduce((obj, [key, value]) => (
//         Object.assign(obj, { [key]: value }) // Be careful! Maps can have non-String keys; object literals can't.
//     ), {});
// }
//
//
// function map(iter, cb) {
//     let out = [];
//     let i = -1;
//     for(let x of iter) {
//         out.push(cb(x,++i));
//     }
//     return out;
// }
//
// function mapValues(iter, cb) {
//     let out = new Array(iter.size);
//     let i = 0;
//     for(let x of iter.values()) {
//         out[i] = cb(x,i);
//         ++i;
//     }
//     return out;
// }
//
// function arrayIncludes(arr, cb) {
//     for(let i=0; i<arr.length; ++i) {
//         if((cb(arr[i],i))) {
//             return true;
//         }
//     }
//     return false;
// }
//
// function filterInPlace(a, condition) {
//     // https://stackoverflow.com/a/37319954/65387
//     let i = 0, j = 0, removed = 0;
//
//     while(i < a.length) {
//         const val = a[i];
//         if(condition(val, i, a)) a[j++] = val;
//         else ++removed;
//         ++i;
//     }
//
//     a.length = j;
//     return removed;
// }
//
// function splitArray(array, index) {
//     return [array.slice(0, index), array.slice(index)];
// }

/**
 * Unwraps a value. If passed a function, evaluates that function with the provided args. Otherwise, returns the value as-is.
 *
 * @param {Function|*} functionOrValue Function or value
 * @param {*} args Arguments to pass if `functionOrValue` is a function
 * @returns {*} The value passed in or the result of calling the function
 */
function resolveValue(functionOrValue, ...args) {
    return typeof functionOrValue === 'function' ? functionOrValue.call(this, ...args) : functionOrValue;
}

function setStateHandler(data, options) {
    this.setState({
        [options.dataProp]: data
    });
}

class MemoryCache {

    constructor() {
        this.cache = LRU({
            max: (window.screen.width < 1080 ? 10 : 30) * 1024 * 1024,
            maxAge: 1000 * 60 * 5,
            length: obj => sizeof(obj)
        });
    }

    get(key) {
        return this.cache.get(key);
    }

    set(key, value) {
        this.cache.set(key, value);
    }

    delete(key) {
        this.cache.del(key);
    }

    clear() {
        this.cache.reset();
    }
}

const ARRAY_ESTIMATE = 10;
const OBJECT_ESTIMATE = 50;

// https://stackoverflow.com/questions/1248302/how-to-get-the-size-of-a-javascript-object
function sizeof(obj) {
    let stack = [obj];
    let bytes = 0;

    while (stack.length) {
        let value = stack.pop();

        bytes += 2; // variable type
        if (value == null || typeof value === 'boolean') {
            bytes += 4;
        } else if (typeof value === 'string') {
            bytes += 8 + value.length * 2;
        } else if (typeof value === 'object') {
            if (Array.isArray(value)) {
                bytes += 8; // length
                if (value.length <= ARRAY_ESTIMATE) {
                    for (let x of value) {
                        stack.push(x);
                    }
                } else {
                    // average out the size so that we don't have to do a deep count
                    let arrSize = 0;
                    for (let i = 0; i < ARRAY_ESTIMATE; ++i) {
                        arrSize += sizeof(value[i]);
                    }
                    bytes += arrSize / ARRAY_ESTIMATE * value.length;
                }
            } else {
                bytes += 12; // meta-data
                let keys = Object.keys(value);
                if (keys.length <= OBJECT_ESTIMATE) {
                    for (let key of keys) {
                        stack.push(key);
                        stack.push(value[key]);
                    }
                } else {
                    let objSize = 0;
                    for (let i = 0; i < OBJECT_ESTIMATE; ++i) {
                        objSize += sizeof(keys[i]) + sizeof(value[keys[i]]);
                    }
                    bytes += objSize / OBJECT_ESTIMATE * keys.length;
                }
            }
        } else {
            bytes += 8;
        }
    }

    return bytes;
}

/*
should have the following interface:

- set
- get
- delete
- clear

similar to "Map", but not quite the same as storage (https://developer.mozilla.org/en-US/docs/Web/API/Storage).


should we use a MinHeap for LRU or.... what? "set associative cache"
MinHeap would allow us to look up the 'oldest' item in O(1) but what about key lookup?
see also https://github.com/epoberezkin/sacjs
 */

export { FetchPolicy, AjaxLoader, MemoryCache };
//# sourceMappingURL=browser.js.map
